{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE TypeOperators #-}

module Control.Monad.IRTree (
  -- * IRTree
  IRTree,
  extract,
  probe,
  reduce,
  reduceAdaptive,

  -- * IRTreeT
  IRTreeT,
  probeT,
  extractT,
  reduceT,
  reduceAdaptiveT,

  -- * helpers
  binsearch,
  expsearch,

  -- * Re-exports
  module Control.Monad.Reader,
  module Data.RPath,
) where

import Control.Monad.RWS.Strict
import Control.Monad.Reader
import Control.Monad.Reduce
import Data.Bits
import Data.Foldable
import Data.Function
import Data.Functor
import Data.Functor.Identity
import Data.RPath
import qualified Data.Sequence as Seq

type IRTree l = IRTreeT l Identity

newtype IRTreeT l m i = IRTreeT (RWST RPath (Endo [l]) Int m i)
  deriving
    (Functor, Applicative, Monad)
    via (RWST RPath (Endo [l]) Int m)

instance (Monad m) => MonadReduce l (IRTreeT l m) where
  check l = IRTreeT . RWST $ \rp i -> do
    pure (indexChoice rp i, i + 1, Endo (l :))
  {-# INLINE check #-}

extract :: IRTree l i -> i
extract t = runIdentity $ extractT t
{-# INLINE extract #-}

extractT :: (Functor m) => IRTreeT l m i -> m i
extractT (IRTreeT m) = fmap (\(i, _, _) -> i) (runRWST m "" 0)
{-# INLINE extractT #-}

probe :: IRTree l i -> RPath -> (i, [AnnotatedChoice l], Int)
probe t pth = runIdentity $ probeT t pth
{-# INLINE probe #-}

{- | Probe the IRTree at a path RPath, and return the input, a list of annotated choices,
and the number of guesses made (can be negative to indicate that not all choices in
the path was used.
-}
probeT :: (Functor m) => IRTreeT l m i -> RPath -> m (i, [AnnotatedChoice l], Int)
probeT (IRTreeT m) pth =
  runRWST m pth 0 <&> \(i, x, l) ->
    (i, annotateChoices pth (appEndo l []), x - numberOfChoices pth)
{-# INLINE probeT #-}

reduce
  :: (Monad m)
  => ([AnnotatedChoice l] -> i -> m Bool)
  -> IRTree l i
  -> m i
reduce = reduceT (pure . runIdentity)
{-# INLINE reduce #-}

-- | Interpreted reduction with an m base monad
reduceT
  :: (Monad m, Functor t)
  => (forall a. t a -> m a)
  -> ([AnnotatedChoice l] -> i -> m Bool)
  -> IRTreeT l t i
  -> m i
reduceT lift_ p rt = do
  (k', _, _) <- _probe Seq.empty
  (\f -> f Seq.empty k') $ fix \rec sq k -> do
    (i, l, left) <- _probe (sq Seq.|> True)
    if left < 0
      then pure k
      else do
        t <- p l i
        rec (sq Seq.|> t) (if t then i else k)
 where
  _probe sq = lift_ . probeT rt . fromChoiceList $ toList sq
{-# INLINE reduceT #-}

reduceAdaptive
  :: (Monad m)
  => ([AnnotatedChoice l] -> i -> m Bool)
  -> IRTree l i
  -> m i
reduceAdaptive = reduceAdaptiveT (pure . runIdentity)
{-# INLINE reduceAdaptive #-}

-- | Interpreted reduction with an m base monad, but using exponential search.
reduceAdaptiveT
  :: (Monad m, Functor t)
  => (forall a. t a -> m a)
  -- ^ a lift of monad m into t (normally @id@ or @lift@)
  -> ([AnnotatedChoice l] -> i -> m Bool)
  -> IRTreeT l t i
  -> m i
reduceAdaptiveT lift_ p rt = do
  Seq.empty & fix \rec !sq -> do
    (d', c) <- expsearch \d -> do
      let sq' = sq <> Seq.replicate d True
      (i, l, left) <- _probe sq'
      t <- p l i
      pure (t, left)
    if c
      then rec (sq <> Seq.replicate d' True <> Seq.singleton False)
      else do
        (i, _, _) <- _probe (sq <> Seq.replicate d' True)
        pure i
 where
  _probe sq = lift_ . probeT rt . fromChoiceList $ toList sq

{- | Do an exponential search until we find failing case, return
the number of successfull cases, and a boolean to indicate if
we should continue the search.
-}
expsearch :: (Monad m) => (Int -> m (Bool, Int)) -> m (Int, Bool)
expsearch p =
  1 & fix \rec !d -> do
    p d >>= \case
      (True, left)
        | left < 0 -> pure (d + left, False)
        | otherwise -> rec (d `shiftL` 1)
      (False, left)
        | left < 0 && d == 1 -> do
            pure (0, False)
        | otherwise -> do
            d' <-
              binsearch
                (fmap (fmap fst) p)
                (d `shiftR` 1)
                (min d (d + left))
            -- continue unless we d is one and left is false
            pure (d', True)

binsearch :: (Monad m) => (Int -> m Bool) -> Int -> Int -> m Int
binsearch p = go
 where
  go !m !h
    | m >= pivot = do
        pure m
    | otherwise = do
        p pivot >>= \case
          True -> go pivot h
          False -> go m pivot
   where
    pivot = (m + h) `shiftR` 1
